from .stage_interface import StageInterface as ImplementsStageInterface
from ..partials.with_base.type_payload import TPayload
from ..partials.with_base.stage import (
    Stage as WithStageBaseConcern,
)
from ..partials.with_stage_as_callable.stage_mixin import (
    StageMixin as WithStageAsCallableConcern,
)


class Stage(
    WithStageAsCallableConcern[TPayload],
    WithStageBaseConcern,
    ImplementsStageInterface[TPayload],
):
    pass



from typing import Any, Protocol

from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as ImplementsPipelineInterface,
)
from ..with_base.type_payload import TPayload


class PipelineMixin(
    ImplementsPipelineInterface[TPayload],
    Protocol[TPayload],
):
    async def __call__(
        self,
        payload: TPayload,
        *args: Any,
        **kwds: Any,
    ) -> TPayload:
        return await self.process(payload, *args, **kwds)



from typing import Awaitable, Callable, Concatenate, ParamSpec

from .type_payload import TPayload

TArgs = ParamSpec("TArgs")  # capture *args and **kwds

PipelineCallable = (
    Callable[Concatenate[TPayload, TArgs], Awaitable[TPayload]]
    | Callable[Concatenate[TPayload, TArgs], TPayload]
)



from typing import Optional, Protocol, Self

from .pipelineable_facade import PipelineableFacade
from ..with_base.type_pipeline_callable import PipelineCallable
from ..with_base.type_payload import TPayload
from ..with_base.pipeline_interface import (
    PipelineInterface as WithPipelineBaseInterface,
)
from ..with_pipeline_as_list.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsListInterface,
)
from ..with_pipeline_processor.processor_interface import ProcessorInterface


class PipelineInterfaceMixin(
    WithPipelineAsListInterface[TPayload],
    WithPipelineBaseInterface,
    Protocol[TPayload],
):
    def build(
        self,
        processor: Optional[ProcessorInterface[TPayload]] = None,
    ) -> PipelineableFacade[TPayload]: ...

    def add(self, item: PipelineCallable[TPayload, ...]) -> Self: ...



from abc import abstractmethod
from typing import Any, Protocol

from ..with_base.type_payload import TPayload
from ..with_base.type_pipeline_callable import PipelineCallable


class ProcessorInterface(
    Protocol[TPayload],
):
    @abstractmethod
    async def process(
        self,
        payload: TPayload,
        stages: list[PipelineCallable[TPayload, ...]],
        *args: Any,
        **kwds: Any,
    ) -> TPayload: ...

    async def _call_stage(
        self,
        payload: TPayload,
        stage: PipelineCallable[TPayload, ...],
        *args: Any,
        **kwds: Any,
    ) -> TPayload: ...



from typing import Protocol

from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as ImplementsPipelineInterface,
)
from ..with_base.type_payload import TPayload
from ..with_base.type_pipeline_callable import PipelineCallable
from ...support.renamable_list.traits.has_immutability import (
    HasImmutability as HasImmutabilityConcern,
)


class PipelineMixin(
    HasImmutabilityConcern[PipelineCallable[TPayload, ...]],
    ImplementsPipelineInterface[TPayload],
    Protocol[TPayload],
):
    pass



from abc import abstractmethod
from typing import Any, Protocol

from ..with_base.stage_interface import (
    StageInterface as WithStageBaseInterface,
)
from ..with_base.type_payload import TPayload


class StageInterfaceMixin(
    WithStageBaseInterface,
    Protocol[TPayload],
):
    @abstractmethod
    async def __call__(
        self,
        payload: TPayload,
        *args: Any,
        **kwds: Any,
    ) -> TPayload: ...



from abc import abstractmethod
from typing import Any, Protocol

from .stage_interface_mixin import (
    StageInterfaceMixin as ImplementsStageInterface,
)
from ..with_base.type_payload import TPayload


class StageMixin(
    ImplementsStageInterface[TPayload],
    Protocol[TPayload],
):
    @abstractmethod
    async def __call__(
        self,
        payload: TPayload,
        *args: Any,
        **kwds: Any,
    ) -> TPayload:
        pass



from typing import Optional, Protocol, Self

from .pipelineable_facade import TPipelineable
from ..with_base.type_pipeline_callable import PipelineCallable
from ..with_base.type_payload import TPayload
from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as ImplementsPipelineInterface,
)
from ..with_pipeline_processor.processor_interface import ProcessorInterface


class PipelineMixin(
    ImplementsPipelineInterface[TPayload],
    Protocol[TPayload, TPipelineable],
):
    pipelineable_class: type[TPipelineable]

    def __init__(self) -> None:
        super().__init__()

        self.set_items(self.get_items() or [])

    def add(self, item: PipelineCallable[TPayload, ...]) -> Self:
        return self.add_item(item)

    def build(
        self,
        processor: Optional[ProcessorInterface[TPayload]] = None,
    ) -> TPipelineable:
        pipelineable = self.pipelineable_class()

        return pipelineable.set_items(
            [*self.get_items()],
        ).set_processor(
            processor or pipelineable.get_processor(),
        )



from typing import Any, Generic

from ..partials.with_base.type_pipeline_callable import PipelineCallable
from ..partials.with_base.type_payload import TPayload
from ..partials.with_pipeline_processor.processor_interface import (
    ProcessorInterface as ImplementsProcessorInterface,
)
from ..partials.with_pipeline_processor.processor import (
    Processor as WithProcessorConcern,
)


class ChainedProcessor(
    WithProcessorConcern[TPayload],
    ImplementsProcessorInterface[TPayload],
    Generic[TPayload],
):
    async def process(
        self,
        payload: TPayload,
        stages: list[PipelineCallable[TPayload, ...]],
        *args: Any,
        **kwds: Any,
    ) -> TPayload:
        for stage in stages:
            payload = await self._call_stage(payload, stage, *args, **kwds)

        return payload



from typing import Protocol

from ..partials.with_base.type_payload import TPayload
from ..partials.with_pipeline_processor.processor_interface import (
    ProcessorInterface as WithProcessorBaseInterface,
)


class ProcessorInterface(
    WithProcessorBaseInterface[TPayload],
    Protocol[TPayload],
):
    pass



from typing import Protocol, Self

from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as ImplementsPipelineInterface,
)
from ..with_base.type_payload import TPayload
from ..with_base.type_pipeline_callable import PipelineCallable
from ...support.renamable_list.renamable_list import (
    RenamableList as RenamableListConcern,
)


class PipelineMixin(
    RenamableListConcern[PipelineCallable[TPayload, ...]],
    ImplementsPipelineInterface[TPayload],
    Protocol[TPayload],
):
    stage_instances: list[PipelineCallable[TPayload, ...]] = []

    def get_items(self) -> list[PipelineCallable[TPayload, ...]]:
        return self.stage_instances

    def set_items(self, items: list[PipelineCallable[TPayload, ...]]) -> Self:
        self.stage_instances = items

        return self



from abc import abstractmethod
import inspect
from typing import Any, Protocol

from .processor_interface import (
    ProcessorInterface as ImplementsProcessorInterface,
)
from ..with_base.type_payload import TPayload
from ..with_base.type_pipeline_callable import PipelineCallable


class Processor(
    ImplementsProcessorInterface[TPayload],
    Protocol[TPayload],
):
    @abstractmethod
    async def process(
        self,
        payload: TPayload,
        stages: list[PipelineCallable[TPayload, ...]],
        *args: Any,
        **kwds: Any,
    ) -> TPayload:
        pass

    async def _call_stage(
        self,
        payload: TPayload,
        stage: PipelineCallable[TPayload, ...],
        *args: Any,
        **kwds: Any,
    ) -> TPayload:
        result = stage(payload, *args, **kwds)

        if inspect.isawaitable(result):
            return await result

        return result



from .processor_interface import (
    ProcessorInterface as ImplementsProcessorInterface,
)
from ..partials.with_base.type_payload import TPayload
from ..partials.with_pipeline_processor.processor import (
    Processor as WithProcessorBaseConcern,
)


class Processor(
    WithProcessorBaseConcern[TPayload],
    ImplementsProcessorInterface[TPayload],
):
    pass



from typing import Any, Generic

from thecodecrate_pipeline.partials.with_base.type_pipeline_callable import (
    PipelineCallable,
)

from .stage_facade import StageFacade
from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as ImplementsPipelineInterface,
)
from ..with_base.type_payload import TPayload

PipelineCallableOrStage = (
    type[StageFacade[TPayload]]
    | StageFacade[TPayload]
    | PipelineCallable[TPayload, ...]
)


class PipelineMixin(
    ImplementsPipelineInterface[TPayload],
    Generic[TPayload],
):
    stages: list[PipelineCallableOrStage[TPayload]] = []

    def __init__(
        self,
        *args: Any,
        **kwds: Any,
    ) -> None:
        super().__init__(*args, **kwds)

        self._instantiate_stages()

    def _instantiate_stages(self) -> None:
        self.stage_instances = [
            stage() if isinstance(stage, type) else stage
            for stage in self.stages
        ]



from typing import Protocol

from ..with_base.type_payload import TPayload
from ..with_base.type_pipeline_callable import PipelineCallable
from ..with_base.pipeline_interface import (
    PipelineInterface as WithPipelineBaseInterface,
)
from ...support.renamable_list.renamable_list_interface import (
    RenamableListInterface,
)


class PipelineInterfaceMixin(
    RenamableListInterface[PipelineCallable[TPayload, ...]],
    WithPipelineBaseInterface,
    Protocol[TPayload],
):
    pass



from typing import Protocol, Self, Optional

from ..with_pipeline_processor.processor_interface import ProcessorInterface
from ..with_base.type_pipeline_callable import PipelineCallable
from ..with_base.type_payload import TPayload
from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as ImplementsPipelineInterface,
)


class PipelineMixin(
    ImplementsPipelineInterface[TPayload],
    Protocol[TPayload],
):
    def __init__(
        self,
        processor: Optional[ProcessorInterface[TPayload]] = None,
    ) -> None:
        super().__init__()

        self.set_items(self.get_items() or []).set_processor(
            processor or self.get_processor()
        )

    def pipe(self, stage: PipelineCallable[TPayload, ...]) -> Self:
        return self.add_item(stage)
